#include <linux/autoconf.h>

#include "assyms.h"

LBITS_040 = LPAGE_PRESENT|LPAGE_CACHE040

.text
	.even
.globl _savectx
_savectx:
	/* first argument is the task_struct */
	movel	sp@(4),a0

	/* save status register */
	movew	sr,a0@(LTS_SR)

	/* save current alternate address space */
	movec	sfc,d0
	movel	d0,a0@(LTS_FS)

	/* save user stack pointer */
	movel	usp,a1
	movel	a1,a0@(LTS_USP)

	/* save non-scratch registers */
	movml	d2-d7/a2-a7,a0@(LTS_REGS)

#ifdef CONFIG_M68K_FLOAT
	/* save floating point state */
	fsave	a0@(LTS_FPSTATE)
	tstw	a0@(LTS_FPSTATE)
	beq	1f	    /* NULL state, no need to save regs */
	fmovem	fp0-fp7,a0@(LTS_FPREGS)
	fmovem	fpiar/fpcr/fpsr,a0@(LTS_FPCNTL)
1:
#endif /* CONFIG_M68K_FLOAT */

	/* second argument is altreturn */
	movel	sp@(8),d0
	beq	Lnotalt

	/* copy return address from this stack to the new one */
	movel	sp,a1
	subl	#LKSTACK_ADDR,a1
	addl	a0@(LTS_KSTACKP),a1
	movel	sp@,a1@
Lnotalt:
	moveq	#0,d0
	rts


.globl _resume
	.even
_resume:
	/* current */
	movel	_current,a0

	/* save sr */
	movew	sr,a0@(LTS_SR)

	/* disable interrupts */
	oriw	#0x0700,sr

	/* save fs (sfc,dfc) (may be pointing to kernel memory) */
	movec	sfc,d0
	movel	d0,a0@(LTS_FS)

	/* save usp */
	movel	usp,a1
	movel	a1,a0@(LTS_USP)

	/* save non-scratch registers */
	moveml	d2-d7/a2-a7,a0@(LTS_REGS)

#ifdef CONFIG_M68K_FLOAT
	/* save floating point state */
	fsave	a0@(LTS_FPSTATE)
	tstw	a0@(LTS_FPSTATE)
	beq	1f	    /* NULL state, no need to save regs */
	fmovemx fp0-fp7,a0@(LTS_FPREGS)
	fmovem	fpiar/fpcr/fpsr,a0@(LTS_FPCNTL)
1:
#endif /* CONFIG_M68K_FLOAT */

	/* switch current task */
	movel	sp@(4),a0
	movel	a0,_current

	/* 68040 ? */
	movel	_sys_info+LSI_CPU,d1
	btst	#2,d1
	bnes	1f

	/*
	 * switch address space
	 */

	/* flush MC68030/MC68020 caches (they are virtually addressed) */
	movec	cacr,d0
	oril	#LFLUSH_I_AND_D,d0
	movec	d0,cacr

	/* first, the kernel stack */
	movel	a0@(LTS_PMAP+LPMAP_KSTACKP),d0
	orib	#LPAGE_PRESENT,d0
	movel	_kstack_descp,a1
	movel	d0,a1@

	/* switch the root pointer */
	pmove	a0@(LTS_PMAP+LPMAP_CRP),crp

	/* flush address translation cache (probably not needed */
	pflusha

	bras	2f	/* skip m68040 stuff */

1:
	/*
	 * switch address space
	 */

	/* first, the kernel stack */
	movel	a0@(LTS_PMAP+LPMAP_KSTACKP),d0
	oriw	#LBITS_040,d0
	movel	_kstack_descp,a1
	movel	d0,a1@

	/* flush address translation cache (user entries) */
	.word	0xf510			/* pflushan */

	/* switch the root pointer */
	movel	a0@(LTS_PMAP+LPMAP_CRP+4),d0
	.long	0x4e7b0806		/* movec d0,urp */


2:
#ifdef CONFIG_M68K_FLOAT
	/* restore floating point unit state */
	tstw	a0@(LTS_FPSTATE)
	beqs	1f	    /* NULL state; no need to load regs */
	fmovemx a0@(LTS_FPREGS),fp0-fp7
	fmovem	a0@(LTS_FPCNTL),fpiar/fpcr/fpsr
1:
	frestore a0@(LTS_FPSTATE)
#endif /* CONFIG_M68K_FLOAT */

	/* restore non-scratch registers */
	moveml	a0@(LTS_REGS),d2-d7/a2-a7

	/* restore usp */
	movel	a0@(LTS_USP),a1
	movel	a1,usp

	/* restore fs (sfc,dfc) */
	movel	a0@(LTS_FS),d0
	movec	d0,sfc
	movec	d0,dfc

	/* restore status register */
	movew	a0@(LTS_SR),sr

	/* return 1 for newly forked task */
	moveq	#1,d0
	rts
